/* Dlist - dynamic list
 * Copyright (C) 2016-2024 Frediano Ziglio
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <assert.h>
#include <freetds/utils/dlist.h>

#if ENABLE_EXTRA_CHECKS
void dlist_ring_check(dlist_ring *ring)
{
	const dlist_ring *item = ring;
	do {
		assert(item->prev->next == item);
		assert(item->next->prev == item);
		item = item->next;
	} while (item != ring);
}
#endif
